from pwnlib.tubes.listen import listen
from netpwn.config import SHELL_STABILIZATION_METHODS


class ListenerService:
    def __init__(self, lport: int):
        self.lport = lport

    def prepare_listener(self) -> listen:
        listener = listen(self.lport)

        return listener

    def is_rev_shell(self, listener: listen):
        data_received = listener.recv(timeout=0.5)
        listener.unrecv(data_received)
        # Check that we got a shell
        if (not b'$' in data_received and not b'Windows' in data_received):
            return False
        return True

    def can_upgrade_shell(self, shell: listen):
        data_received = shell.recv(timeout=0.5)
        shell.unrecv(data_received)

        if (not b'Windows' in data_received):
            return True
        return False

    def upgrade_shell(self, shell: listen):
        shell.sendline(b'export HISTFILE=/dev/null') # Avoid history
        # Get pty
        shell.sendline(f'{self.find_pty_spawn_vector(shell)}; exit'.encode())

        shell.recv(timeout = None) # Wait for pty to spawn
        shell.sendline(b'export HISTFILE=/dev/null')
        shell.sendline(b'stty rows 38 columns 116')
        shell.sendline(b""" alias ls='ls --color=auto'""")
        shell.sendline(b'export TERM=xterm')
        shell.sendline(b'history -c')
        shell.clean(timeout=1)
    
    def find_pty_spawn_vector(self, shell: listen) -> str:
        shell.clean()

        for binary in SHELL_STABILIZATION_METHODS:
            shell.sendline(f'which {binary}'.encode())
            result = shell.recvline(timeout=2).decode()
            if(binary in result and 'not found' not in result):
                for shell_binary in SHELL_STABILIZATION_METHODS[binary]:
                    shell.sendline(f'which {shell_binary}'.encode())
                    result = shell.recvline(timeout=2).decode()
                    if(shell_binary in result and 'not found' not in result):
                        return SHELL_STABILIZATION_METHODS[binary][shell_binary]
            shell.clean()
        
        raise Exception('No pty spawn vector found')



VERSION = '1.1'

SHELL_STABILIZATION_METHODS = {
    'python': {
        'bash': """python -c 'import pty; pty.spawn("/bin/bash")'""",
        'sh': """python -c 'import pty; pty.spawn("/bin/bash")'"""
    },
    'python3': {
        'bash': """python3 -c 'import pty; pty.spawn("/bin/bash")'""",
        'sh': """python3 -c 'import pty; pty.spawn("/bin/sh")'"""
    },
    'script': {
        'bash': 'script -qc /bin/bash /dev/null',
        'sh': 'script -qc /bin/sh /dev/null'
    }
}



import sys, termios, tty


class TerminalService:
    def __init__(self):
        self.initial_options = termios.tcgetattr(sys.stdin)
        self.pty = False

    def restore_tty(self):
    	termios.tcsetattr(sys.stdin, termios.TCSANOW, self.initial_options)

    def set_raw_mode(self):
    	tty.setraw(sys.stdin.fileno())



import argparse
from netpwn.config import VERSION


class ParametersParserService:
	def __init__(self):
		self.parser = argparse.ArgumentParser(description='A listener capable of stabilize Linux and MacOS reverse shells automatically')
		self.parser.add_argument('-v', '--version', action='version', version='%(prog)s version ' + VERSION)
		self.parser.add_argument('--no-pty', action='store_true', help='if this flag is set, no shell stabilization is perform')
		self.parser.add_argument('-P', '--lport', type=int, help='the port used to listen for the reverse shell (Default: 8080)', default=8080)

	def parse_params(self) -> argparse.Namespace:
		return self.parser.parse_args()



import unittest
from netpwn.services import ListenerService
from pwnlib.tubes.remote import remote
from netpwn.config import SHELL_STABILIZATION_METHODS


class ListenerServiceTest(unittest.TestCase):
    lhost = '127.0.0.1'
    lport = 8080
    
    def test_prepare_listener(self):
        listener_service = ListenerService(self.lport)
        listener = listener_service.prepare_listener()
        remote(self.lhost, self.lport)
        
        self.assertTrue(listener.connected())
        listener.close()

    def test_can_upgrade_shell(self):
        listener_service = ListenerService(self.lport)
        listener = listener_service.prepare_listener()
        remote_shell = remote(self.lhost, self.lport)
        remote_shell.sendline(b'$')

        self.assertTrue(listener_service.can_upgrade_shell(listener))

    def test_cannot_upgrade_shell(self):
        listener_service = ListenerService(self.lport)
        listener = listener_service.prepare_listener()
        remote_shell = remote(self.lhost, self.lport)
        remote_shell.sendline(b'Microsoft Windows')

        self.assertFalse(listener_service.can_upgrade_shell(listener))
    
    def test_upgrade_shell(self):
        listener_service = ListenerService(self.lport)
        listener = listener_service.prepare_listener()
        remote_shell = remote(self.lhost, self.lport)
        expected_commands = ['export HISTFILE=/dev/null', 'which python', 'which bash', 'python -c \'import pty; pty.spawn("/bin/bash")\'; exit', 'export HISTFILE=/dev/null', 'stty rows 38 columns 116', " alias ls='ls --color=auto'", 'export TERM=xterm', 'history -c']
        commands_executed = []

        # Mock listener a bit
        def mocked_sendline(x): 
            remote_shell.sendline(b'fake@fake:~$ /usr/bin/python /bin/bash')
            commands_executed.append(x.decode())
        listener.sendline = mocked_sendline

        listener_service.upgrade_shell(listener)
        listener.close()
        self.assertListEqual(commands_executed, expected_commands)
    
    def test_find_pty_spawn_vector(self):
        listener_service = ListenerService(self.lport)
        listener = listener_service.prepare_listener()
        remote_shell = remote(self.lhost, self.lport)

        # Mock listener a bit
        def mocked_sendline(x): remote_shell.sendline(b'fake@fake:~$ /usr/bin/python /bin/bash')
        listener.sendline = mocked_sendline

        self.assertEqual(listener_service.find_pty_spawn_vector(listener), SHELL_STABILIZATION_METHODS['python']['bash'])
        listener.close()
    
    def test_find_pty_spawn_vector_error(self):
        listener_service = ListenerService(self.lport)
        listener = listener_service.prepare_listener()
        remote_shell = remote(self.lhost, self.lport)

        # Mock listener a bit
        def mocked_sendline(x): remote_shell.sendline(b'fake@fake:~$ ')
        listener.sendline = mocked_sendline

        self.assertRaises(Exception, listener_service.find_pty_spawn_vector, listener)
        listener.close()


if __name__ == '__main__':
    unittest.main()



import unittest, sys
import unittest.mock
from argparse import Namespace
from io import StringIO
from netpwn.services import ParametersParserService


class ParametersParserServiceTest(unittest.TestCase):
    def test_arguments_parse(self):
        sys.argv = ['./netpawn.py', '--lport', '8000']
        parameterParserService = ParametersParserService()

        with unittest.mock.patch('sys.stderr', new=StringIO()):
            try:
                params = parameterParserService.parse_params()
                self.assertEqual(params, Namespace(lport=8000, no_pty=False))
            except SystemExit:
                raise Exception('Wrong parameters!')

    def test_arguments_parse_default_port(self):
        sys.argv = ['./netpawn.py']
        parameterParserService = ParametersParserService()

        with unittest.mock.patch('sys.stderr', new=StringIO()):
            try:
                params = parameterParserService.parse_params()
                self.assertEqual(params, Namespace(lport=8080, no_pty=False))
            except SystemExit:
                raise Exception('Wrong parameters!')

    def test_arguments_parse_error(self):
        sys.argv = ['./netpawn.py', '--lport']
        parameterParserService = ParametersParserService()

        with unittest.mock.patch('sys.stderr', new=StringIO()) as fakeOutput:
            try:
                parameterParserService.parse_params()
            except SystemExit:
                self.assertTrue('error: argument -P/--lport: expected one argument' in fakeOutput.getvalue().strip())


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

setup(
    name='netpwn',
    version='1.1.post2',
    license='GPL-2.0',
    description='A netcat listener alternative with automatic shell stabilization.',
    author='Ángel Heredia',
    packages=find_packages(),
    url='https://github.com/anthares101/netpwn',
    download_url="https://github.com/anthares101/netpwn/releases",
    keywords='windows macos linux shell reverse-shell tool hacking netcat tty pty cybersecurity reverse pwntools hacktoberfest kali',
    python_requires='>=3',
    install_requires=[
        'pwntools',
    ],
    entry_points='''
        [console_scripts]
        netpwn=netpwn.__main__:main
    ''',
)



#!/usr/bin/env python3

import os, threading
from argparse import Namespace
from pwnlib.log import install_default_handler, getLogger
from pwnlib.tubes.listen import listen
from netpwn.services import (
    ListenerService,
    ParametersParserService,
    TerminalService
)


# Init pwntools stuff and safe terminal conf
terminal_service = TerminalService()
install_default_handler()
log = getLogger('pwnlib')

def handler(listener: listen):
    while listener.connected('recv'): pass
    listener.shutdown('send')

    # Restore terminal conf and exit
    terminal_service.restore_tty()

    log.setLevel(1)
    log.info(f'Closed connection to {listener.rhost} port {listener.rport}')
    os._exit(0)


def netpwn(args: Namespace):
    lport = args.lport
    no_pty = args.no_pty

    listener_service = ListenerService(lport)
    listener = listener_service.prepare_listener()
    listener.wait_for_connection()

    if(listener_service.is_rev_shell(listener) and listener_service.can_upgrade_shell(listener) and not no_pty):
        with log.progress('Trying to stabilize the shell...') as p:
            try:
                listener_service.upgrade_shell(listener)
                terminal_service.pty = True
                p.success('Shell stabilized!')
            except:
                p.failure('Shell stabilization not possible, a non pty shell will be provided')
           
    # Start handler to control the closing
    handler_thread = threading.Thread(target=handler, args=(listener,))
    handler_thread.setDaemon(True)
    handler_thread.start()
    
    # Start interactive mode
    log.info('Switching to interactive mode')
    log.setLevel('error')
    if(terminal_service.pty):
        terminal_service.set_raw_mode()
        listener.sendline(b'') # Make the prompt appear
    listener.interactive()
    
def main():
    print('Netpwn  - A netcat listener alternative\n')

    parameterParserService = ParametersParserService()
    args = parameterParserService.parse_params()

    try:
        netpwn(args)
    except KeyboardInterrupt:
        log.failure('Interrupted')
    except Exception as error:
        log.failure(error.__str__())

if __name__ == '__main__':
    main()
